@@ -25,11 +25,20 @@ namespace ts {
             parenthesizeExpressionForDisallowedComma,
             parenthesizeExpressionOfExpressionStatement,
             parenthesizeConciseBodyOfArrowFunction,
-            parenthesizeMemberOfConditionalType,
-            parenthesizeMemberOfElementType,
-            parenthesizeElementTypeOfArrayType,
-            parenthesizeConstituentTypesOfUnionOrIntersectionType,
+            parenthesizeCheckTypeOfConditionalType,
+            parenthesizeExtendsTypeOfConditionalType,
+            parenthesizeConstituentTypesOfUnionType,
+            parenthesizeConstituentTypeOfUnionType,
+            parenthesizeConstituentTypesOfIntersectionType,
+            parenthesizeConstituentTypeOfIntersectionType,
+            parenthesizeOperandOfTypeOperator,
+            parenthesizeOperandOfReadonlyTypeOperator,
+            parenthesizeNonArrayTypeOfPostfixType,
+            parenthesizeElementTypesOfTupleType,
+            parenthesizeElementTypeOfTupleType,
+            parenthesizeTypeOfOptionalType,
             parenthesizeTypeArguments,
+            parenthesizeLeadingTypeArgument,
         };
 
         function getParenthesizeLeftSideOfBinaryForOperator(operatorKind: BinaryOperator) {
@@ -310,7 +319,7 @@ namespace ts {
          * Wraps an expression in parentheses if it is needed in order to use the expression for
          * property or element access.
          */
-        function parenthesizeLeftSideOfAccess(expression: Expression): LeftHandSideExpression {
+        function parenthesizeLeftSideOfAccess(expression: Expression, optionalChain?: boolean): LeftHandSideExpression {
             // isLeftHandSideExpression is almost the correct criterion for when it is not necessary
             // to parenthesize the expression before a dot. The known exception is:
             //
@@ -319,7 +328,8 @@ namespace ts {
             //
             const emittedExpression = skipPartiallyEmittedExpressions(expression);
             if (isLeftHandSideExpression(emittedExpression)
-                && (emittedExpression.kind !== SyntaxKind.NewExpression || (emittedExpression as NewExpression).arguments)) {
+                && (emittedExpression.kind !== SyntaxKind.NewExpression || (emittedExpression as NewExpression).arguments)
+                && (optionalChain || !isOptionalChain(emittedExpression))) {
                 // TODO(rbuckton): Verify whether this assertion holds.
                 return expression as LeftHandSideExpression;
             }
@@ -388,38 +398,199 @@ namespace ts {
             return body;
         }
 
-        function parenthesizeMemberOfConditionalType(member: TypeNode): TypeNode {
-            return member.kind === SyntaxKind.ConditionalType ? factory.createParenthesizedType(member) : member;
-        }
+        // Type[Extends] :
+        //     FunctionOrConstructorType
+        //     ConditionalType[?Extends]
 
-        function parenthesizeMemberOfElementType(member: TypeNode): TypeNode {
-            switch (member.kind) {
-                case SyntaxKind.UnionType:
-                case SyntaxKind.IntersectionType:
+        // ConditionalType[Extends] :
+        //     UnionType[?Extends]
+        //     [~Extends] UnionType[~Extends] `extends` Type[+Extends] `?` Type[~Extends] `:` Type[~Extends]
+        //
+        // - The check type (the `UnionType`, above) does not allow function, constructor, or conditional types (they must be parenthesized)
+        // - The extends type (the first `Type`, above) does not allow conditional types (they must be parenthesized). Function and constructor types are fine.
+        // - The true and false branch types (the second and third `Type` non-terminals, above) allow any type
+        function parenthesizeCheckTypeOfConditionalType(checkType: TypeNode): TypeNode {
+            switch (checkType.kind) {
                 case SyntaxKind.FunctionType:
                 case SyntaxKind.ConstructorType:
-                    return factory.createParenthesizedType(member);
+                case SyntaxKind.ConditionalType:
+                    return factory.createParenthesizedType(checkType);
             }
-            return parenthesizeMemberOfConditionalType(member);
+            return checkType;
         }
 
-        function parenthesizeElementTypeOfArrayType(member: TypeNode): TypeNode {
-            switch (member.kind) {
-                case SyntaxKind.TypeQuery:
+        function parenthesizeExtendsTypeOfConditionalType(extendsType: TypeNode): TypeNode {
+            switch (extendsType.kind) {
+                case SyntaxKind.ConditionalType:
+                    return factory.createParenthesizedType(extendsType);
+            }
+            return extendsType;
+        }
+
+        // UnionType[Extends] :
+        //     `|`? IntersectionType[?Extends]
+        //     UnionType[?Extends] `|` IntersectionType[?Extends]
+        //
+        // - A union type constituent has the same precedence as the check type of a conditional type
+        function parenthesizeConstituentTypeOfUnionType(type: TypeNode) {
+            switch (type.kind) {
+                case SyntaxKind.UnionType: // Not strictly necessary, but a union containing a union should have been flattened
+                case SyntaxKind.IntersectionType: // Not strictly necessary, but makes generated output more readable and avoids breaks in DT tests
+                    return factory.createParenthesizedType(type);
+            }
+            return parenthesizeCheckTypeOfConditionalType(type);
+        }
+
+        function parenthesizeConstituentTypesOfUnionType(members: readonly TypeNode[]): NodeArray<TypeNode> {
+            return factory.createNodeArray(sameMap(members, parenthesizeConstituentTypeOfUnionType));
+        }
+
+        // IntersectionType[Extends] :
+        //     `&`? TypeOperator[?Extends]
+        //     IntersectionType[?Extends] `&` TypeOperator[?Extends]
+        //
+        // - An intersection type constituent does not allow function, constructor, conditional, or union types (they must be parenthesized)
+        function parenthesizeConstituentTypeOfIntersectionType(type: TypeNode) {
+            switch (type.kind) {
+                case SyntaxKind.UnionType:
+                case SyntaxKind.IntersectionType: // Not strictly necessary, but an intersection containing an intersection should have been flattened
+                    return factory.createParenthesizedType(type);
+            }
+            return parenthesizeConstituentTypeOfUnionType(type);
+        }
+
+        function parenthesizeConstituentTypesOfIntersectionType(members: readonly TypeNode[]): NodeArray<TypeNode> {
+            return factory.createNodeArray(sameMap(members, parenthesizeConstituentTypeOfIntersectionType));
+        }
+
+        // TypeOperator[Extends] :
+        //     PostfixType
+        //     InferType[?Extends]
+        //     `keyof` TypeOperator[?Extends]
+        //     `unique` TypeOperator[?Extends]
+        //     `readonly` TypeOperator[?Extends]
+        //
+        function parenthesizeOperandOfTypeOperator(type: TypeNode) {
+            switch (type.kind) {
+                case SyntaxKind.IntersectionType:
+                    return factory.createParenthesizedType(type);
+            }
+            return parenthesizeConstituentTypeOfIntersectionType(type);
+        }
+
+        function parenthesizeOperandOfReadonlyTypeOperator(type: TypeNode) {
+            switch (type.kind) {
                 case SyntaxKind.TypeOperator:
-                case SyntaxKind.InferType:
-                    return factory.createParenthesizedType(member);
+                    return factory.createParenthesizedType(type);
             }
-            return parenthesizeMemberOfElementType(member);
+            return parenthesizeOperandOfTypeOperator(type);
         }
 
-        function parenthesizeConstituentTypesOfUnionOrIntersectionType(members: readonly TypeNode[]): NodeArray<TypeNode> {
-            return factory.createNodeArray(sameMap(members, parenthesizeMemberOfElementType));
+        // PostfixType :
+        //     NonArrayType
+        //     NonArrayType [no LineTerminator here] `!` // JSDoc
+        //     NonArrayType [no LineTerminator here] `?` // JSDoc
+        //     IndexedAccessType
+        //     ArrayType
+        //
+        // IndexedAccessType :
+        //     NonArrayType `[` Type[~Extends] `]`
+        //
+        // ArrayType :
+        //     NonArrayType `[` `]`
+        //
+        function parenthesizeNonArrayTypeOfPostfixType(type: TypeNode) {
+            switch (type.kind) {
+                case SyntaxKind.InferType:
+                case SyntaxKind.TypeOperator:
+                case SyntaxKind.TypeQuery: // Not strictly necessary, but makes generated output more readable and avoids breaks in DT tests
+                    return factory.createParenthesizedType(type);
+            }
+            return parenthesizeOperandOfTypeOperator(type);
+        }
 
+        // TupleType :
+        //     `[` Elision? `]`
+        //     `[` NamedTupleElementTypes `]`
+        //     `[` NamedTupleElementTypes `,` Elision? `]`
+        //     `[` TupleElementTypes `]`
+        //     `[` TupleElementTypes `,` Elision? `]`
+        //
+        // NamedTupleElementTypes :
+        //     Elision? NamedTupleMember
+        //     NamedTupleElementTypes `,` Elision? NamedTupleMember
+        //
+        // NamedTupleMember :
+        //     Identifier `?`? `:` Type[~Extends]
+        //     `...` Identifier `:` Type[~Extends]
+        //
+        // TupleElementTypes :
+        //     Elision? TupleElementType
+        //     TupleElementTypes `,` Elision? TupleElementType
+        //
+        // TupleElementType :
+        //     Type[~Extends] // NOTE: Needs cover grammar to disallow JSDoc postfix-optional
+        //     OptionalType
+        //     RestType
+        //
+        // OptionalType :
+        //     Type[~Extends] `?` // NOTE: Needs cover grammar to disallow JSDoc postfix-optional
+        //
+        // RestType :
+        //     `...` Type[~Extends]
+        //
+        function parenthesizeElementTypesOfTupleType(types: readonly (TypeNode | NamedTupleMember)[]): NodeArray<TypeNode> {
+            return factory.createNodeArray(sameMap(types, parenthesizeElementTypeOfTupleType));
+        }
+
+        function parenthesizeElementTypeOfTupleType(type: TypeNode | NamedTupleMember): TypeNode {
+            if (hasJSDocPostfixQuestion(type)) return factory.createParenthesizedType(type);
+            return type;
+        }
+
+        function hasJSDocPostfixQuestion(type: TypeNode | NamedTupleMember): boolean {
+            if (isJSDocNullableType(type)) return type.postfix;
+            if (isNamedTupleMember(type)) return hasJSDocPostfixQuestion(type.type);
+            if (isFunctionTypeNode(type) || isConstructorTypeNode(type) || isTypeOperatorNode(type)) return hasJSDocPostfixQuestion(type.type);
+            if (isConditionalTypeNode(type)) return hasJSDocPostfixQuestion(type.falseType);
+            if (isUnionTypeNode(type)) return hasJSDocPostfixQuestion(last(type.types));
+            if (isIntersectionTypeNode(type)) return hasJSDocPostfixQuestion(last(type.types));
+            if (isInferTypeNode(type)) return !!type.typeParameter.constraint && hasJSDocPostfixQuestion(type.typeParameter.constraint);
+            return false;
+        }
+
+        function parenthesizeTypeOfOptionalType(type: TypeNode): TypeNode {
+            if (hasJSDocPostfixQuestion(type)) return factory.createParenthesizedType(type);
+            return parenthesizeNonArrayTypeOfPostfixType(type);
+        }
+
+        // function parenthesizeMemberOfElementType(member: TypeNode): TypeNode {
+        //     switch (member.kind) {
+        //         case SyntaxKind.UnionType:
+        //         case SyntaxKind.IntersectionType:
+        //         case SyntaxKind.FunctionType:
+        //         case SyntaxKind.ConstructorType:
+        //             return factory.createParenthesizedType(member);
+        //     }
+        //     return parenthesizeMemberOfConditionalType(member);
+        // }
+
+        // function parenthesizeElementTypeOfArrayType(member: TypeNode): TypeNode {
+        //     switch (member.kind) {
+        //         case SyntaxKind.TypeQuery:
+        //         case SyntaxKind.TypeOperator:
+        //         case SyntaxKind.InferType:
+        //             return factory.createParenthesizedType(member);
+        //     }
+        //     return parenthesizeMemberOfElementType(member);
+        // }
+
+        function parenthesizeLeadingTypeArgument(node: TypeNode) {
+            return isFunctionOrConstructorTypeNode(node) && node.typeParameters ? factory.createParenthesizedType(node) : node;
         }
 
         function parenthesizeOrdinalTypeArgument(node: TypeNode, i: number) {
-            return i === 0 && isFunctionOrConstructorTypeNode(node) && node.typeParameters ? factory.createParenthesizedType(node) : node;
+            return i === 0 ? parenthesizeLeadingTypeArgument(node) : node;
         }
 
         function parenthesizeTypeArguments(typeArguments: NodeArray<TypeNode> | undefined): NodeArray<TypeNode> | undefined {
@@ -446,10 +617,19 @@ namespace ts {
         parenthesizeExpressionForDisallowedComma: identity,
         parenthesizeExpressionOfExpressionStatement: identity,
         parenthesizeConciseBodyOfArrowFunction: identity,
-        parenthesizeMemberOfConditionalType: identity,
-        parenthesizeMemberOfElementType: identity,
-        parenthesizeElementTypeOfArrayType: identity,
-        parenthesizeConstituentTypesOfUnionOrIntersectionType: nodes => cast(nodes, isNodeArray),
+        parenthesizeCheckTypeOfConditionalType: identity,
+        parenthesizeExtendsTypeOfConditionalType: identity,
+        parenthesizeConstituentTypesOfUnionType: nodes => cast(nodes, isNodeArray),
+        parenthesizeConstituentTypeOfUnionType: identity,
+        parenthesizeConstituentTypesOfIntersectionType: nodes => cast(nodes, isNodeArray),
+        parenthesizeConstituentTypeOfIntersectionType: identity,
+        parenthesizeOperandOfTypeOperator: identity,
+        parenthesizeOperandOfReadonlyTypeOperator: identity,
+        parenthesizeNonArrayTypeOfPostfixType: identity,
+        parenthesizeElementTypesOfTupleType: nodes => cast(nodes, isNodeArray),
+        parenthesizeElementTypeOfTupleType: identity,
+        parenthesizeTypeOfOptionalType: identity,
         parenthesizeTypeArguments: nodes => nodes && cast(nodes, isNodeArray),
+        parenthesizeLeadingTypeArgument: identity,
     };
 }
