@@ -1,11 +1,11 @@
-namespace findUpDir {
-    import { join, resolve, dirname } from "path";
-    import { existsSync } from "fs";
+namespace Utils {
+    const { join, resolve, dirname } = require("path") as typeof import("path");
+    const { existsSync } = require("fs") as typeof import("fs");
 
     // search directories upward to avoid hard-wired paths based on the
     // build tree (same as scripts/build/findUpDir.js)
 
-    export function findUpFile(name: string) {
+    export function findUpFile(name: string): string {
         let dir = __dirname;
         while (true) {
             const fullPath = join(dir, name);
