@@ -408,9 +408,7 @@ namespace ts.codefix {
                 const importSpecifiers = mapAllOrFail(name.elements, e =>
                     e.dotDotDotToken || e.initializer || e.propertyName && !isIdentifier(e.propertyName) || !isIdentifier(e.name)
                         ? undefined
-                        // (TODO: GH#18217)
-                        // eslint-disable-next-line @typescript-eslint/no-unnecessary-type-assertion
-                        : makeImportSpecifier(e.propertyName && (e.propertyName as Identifier).text, e.name.text));
+                        : makeImportSpecifier(e.propertyName && e.propertyName.text, e.name.text));
                 if (importSpecifiers) {
                     return convertedImports([makeImport(/*name*/ undefined, importSpecifiers, moduleSpecifier, quotePreference)]);
                 }
@@ -546,7 +544,6 @@ namespace ts.codefix {
 
     function functionExpressionToDeclaration(name: string | undefined, additionalModifiers: readonly Modifier[], fn: FunctionExpression | ArrowFunction | MethodDeclaration, useSitesToUnqualify: ESMap<Node, Node> | undefined): FunctionDeclaration {
         return factory.createFunctionDeclaration(
-            getSynthesizedDeepClones(fn.decorators), // TODO: GH#19915 Don't think this is even legal.
             concatenate(additionalModifiers, getSynthesizedDeepClones(fn.modifiers)),
             getSynthesizedDeepClone(fn.asteriskToken),
             name,
@@ -558,7 +555,6 @@ namespace ts.codefix {
 
     function classExpressionToDeclaration(name: string | undefined, additionalModifiers: readonly Modifier[], cls: ClassExpression, useSitesToUnqualify: ESMap<Node, Node> | undefined): ClassDeclaration {
         return factory.createClassDeclaration(
-            getSynthesizedDeepClones(cls.decorators), // TODO: GH#19915 Don't think this is even legal.
             concatenate(additionalModifiers, getSynthesizedDeepClones(cls.modifiers)),
             name,
             getSynthesizedDeepClones(cls.typeParameters),
@@ -586,7 +582,6 @@ namespace ts.codefix {
 
     function makeExportDeclaration(exportSpecifiers: ExportSpecifier[] | undefined, moduleSpecifier?: string): ExportDeclaration {
         return factory.createExportDeclaration(
-            /*decorators*/ undefined,
             /*modifiers*/ undefined,
             /*isTypeOnly*/ false,
             exportSpecifiers && factory.createNamedExports(exportSpecifiers),
