@@ -46,7 +46,7 @@ namespace ts.codefix {
             const { program, fixId } = context;
             const checker = program.getTypeChecker();
             const seen = new Map<string, true>();
-            const typeDeclToMembers = new Map<ClassOrInterface, ClassOrInterfaceInfo[]>();
+            const typeDeclToMembers = new Map<ClassLikeDeclaration | InterfaceDeclaration | TypeLiteralNode, TypeLikeDeclarationInfo[]>();
 
             return createCombinedCodeActions(textChanges.ChangeTracker.with(context, changes => {
                 eachDiagnostic(context, errorCodes, diag => {
@@ -68,7 +68,7 @@ namespace ts.codefix {
                         if (info.kind === InfoKind.Enum) {
                             addEnumMemberDeclaration(changes, checker, info);
                         }
-                        if (info.kind === InfoKind.ClassOrInterface) {
+                        if (info.kind === InfoKind.TypeLikeDeclaration) {
                             const { parentDeclaration, token } = info;
                             const infos = getOrUpdate(typeDeclToMembers, parentDeclaration, () => []);
                             if (!infos.some(i => i.token.text === token.text)) {
@@ -78,11 +78,11 @@ namespace ts.codefix {
                     }
                 });
 
-                typeDeclToMembers.forEach((infos, classDeclaration) => {
-                    const supers = getAllSupers(classDeclaration, checker);
+                typeDeclToMembers.forEach((infos, declaration) => {
+                    const supers = isTypeLiteralNode(declaration) ? undefined : getAllSupers(declaration, checker);
                     for (const info of infos) {
                         // If some superclass added this property, don't add it again.
-                        if (supers.some(superClassOrInterface => {
+                        if (supers?.some(superClassOrInterface => {
                             const superInfos = typeDeclToMembers.get(superClassOrInterface);
                             return !!superInfos && superInfos.some(({ token }) => token.text === info.token.text);
                         })) continue;
@@ -93,11 +93,11 @@ namespace ts.codefix {
                             addMethodDeclaration(context, changes, call, token, modifierFlags & ModifierFlags.Static, parentDeclaration, declSourceFile);
                         }
                         else {
-                            if (isJSFile && !isInterfaceDeclaration(parentDeclaration)) {
+                            if (isJSFile && !isInterfaceDeclaration(parentDeclaration) && !isTypeLiteralNode(parentDeclaration)) {
                                 addMissingMemberInJs(changes, declSourceFile, parentDeclaration, token, !!(modifierFlags & ModifierFlags.Static));
                             }
                             else {
-                                const typeNode = getTypeNode(program.getTypeChecker(), parentDeclaration, token);
+                                const typeNode = getTypeNode(checker, parentDeclaration, token);
                                 addPropertyDeclaration(changes, declSourceFile, parentDeclaration, token.text, typeNode, modifierFlags & ModifierFlags.Static);
                             }
                         }
@@ -107,8 +107,8 @@ namespace ts.codefix {
         },
     });
 
-    const enum InfoKind { Enum, ClassOrInterface, Function, ObjectLiteral, JsxAttributes }
-    type Info = EnumInfo | ClassOrInterfaceInfo | FunctionInfo | ObjectLiteralInfo | JsxAttributesInfo;
+    const enum InfoKind { TypeLikeDeclaration, Enum, Function, ObjectLiteral, JsxAttributes }
+    type Info = TypeLikeDeclarationInfo | EnumInfo | FunctionInfo | ObjectLiteralInfo | JsxAttributesInfo;
 
     interface EnumInfo {
         readonly kind: InfoKind.Enum;
@@ -116,12 +116,12 @@ namespace ts.codefix {
         readonly parentDeclaration: EnumDeclaration;
     }
 
-    interface ClassOrInterfaceInfo {
-        readonly kind: InfoKind.ClassOrInterface;
+    interface TypeLikeDeclarationInfo {
+        readonly kind: InfoKind.TypeLikeDeclaration;
         readonly call: CallExpression | undefined;
         readonly token: Identifier | PrivateIdentifier;
         readonly modifierFlags: ModifierFlags;
-        readonly parentDeclaration: ClassOrInterface;
+        readonly parentDeclaration: ClassLikeDeclaration | InterfaceDeclaration | TypeLiteralNode;
         readonly declSourceFile: SourceFile;
         readonly isJSFile: boolean;
     }
@@ -169,9 +169,9 @@ namespace ts.codefix {
             const param = signature.parameters[argIndex].valueDeclaration;
             if (!(param && isParameter(param) && isIdentifier(param.name))) return undefined;
 
-            const properties = arrayFrom(checker.getUnmatchedProperties(checker.getTypeAtLocation(parent), checker.getTypeAtLocation(param), /* requireOptionalProperties */ false, /* matchDiscriminantProperties */ false));
+            const properties = arrayFrom(checker.getUnmatchedProperties(checker.getTypeAtLocation(parent), checker.getParameterType(signature, argIndex), /* requireOptionalProperties */ false, /* matchDiscriminantProperties */ false));
             if (!length(properties)) return undefined;
-            return { kind: InfoKind.ObjectLiteral, token: param.name, properties, indentation: 0, parentDeclaration: parent };
+            return { kind: InfoKind.ObjectLiteral, token: param.name, properties, parentDeclaration: parent };
         }
 
         if (!isMemberName(token)) return undefined;
@@ -179,16 +179,18 @@ namespace ts.codefix {
         if (isIdentifier(token) && hasInitializer(parent) && parent.initializer && isObjectLiteralExpression(parent.initializer)) {
             const properties = arrayFrom(checker.getUnmatchedProperties(checker.getTypeAtLocation(parent.initializer), checker.getTypeAtLocation(token), /* requireOptionalProperties */ false, /* matchDiscriminantProperties */ false));
             if (!length(properties)) return undefined;
-            return { kind: InfoKind.ObjectLiteral, token, properties, indentation: undefined, parentDeclaration: parent.initializer };
+
+            return { kind: InfoKind.ObjectLiteral, token, properties, parentDeclaration: parent.initializer };
         }
 
         if (isIdentifier(token) && isJsxOpeningLikeElement(token.parent)) {
-            const attributes = getUnmatchedAttributes(checker, token.parent);
+            const target = getEmitScriptTarget(program.getCompilerOptions());
+            const attributes = getUnmatchedAttributes(checker, target, token.parent);
             if (!length(attributes)) return undefined;
             return { kind: InfoKind.JsxAttributes, token, attributes, parentDeclaration: token.parent };
         }
 
-        if (isIdentifier(token) && isCallExpression(parent)) {
+        if (isIdentifier(token) && isCallExpression(parent) && parent.expression === token) {
             return { kind: InfoKind.Function, token, call: parent, sourceFile, modifierFlags: ModifierFlags.None, parentDeclaration: sourceFile };
         }
 
@@ -218,22 +220,24 @@ namespace ts.codefix {
         if (!classDeclaration && isPrivateIdentifier(token)) return undefined;
 
         // Prefer to change the class instead of the interface if they are merged
-        const classOrInterface = classDeclaration || find(symbol.declarations, isInterfaceDeclaration);
-        if (classOrInterface && !isSourceFileFromLibrary(program, classOrInterface.getSourceFile())) {
-            const makeStatic = ((leftExpressionType as TypeReference).target || leftExpressionType) !== checker.getDeclaredTypeOfSymbol(symbol);
-            if (makeStatic && (isPrivateIdentifier(token) || isInterfaceDeclaration(classOrInterface))) return undefined;
+        const declaration = classDeclaration || find(symbol.declarations, d => isInterfaceDeclaration(d) || isTypeLiteralNode(d)) as InterfaceDeclaration | TypeLiteralNode | undefined;
+        if (declaration && !isSourceFileFromLibrary(program, declaration.getSourceFile())) {
+            const makeStatic = !isTypeLiteralNode(declaration) && ((leftExpressionType as TypeReference).target || leftExpressionType) !== checker.getDeclaredTypeOfSymbol(symbol);
+            if (makeStatic && (isPrivateIdentifier(token) || isInterfaceDeclaration(declaration))) return undefined;
 
-            const declSourceFile = classOrInterface.getSourceFile();
-            const modifierFlags = (makeStatic ? ModifierFlags.Static : 0) | (startsWithUnderscore(token.text) ? ModifierFlags.Private : 0);
+            const declSourceFile = declaration.getSourceFile();
+            const modifierFlags = isTypeLiteralNode(declaration) ? ModifierFlags.None :
+                (makeStatic ? ModifierFlags.Static : ModifierFlags.None) | (startsWithUnderscore(token.text) ? ModifierFlags.Private : ModifierFlags.None);
             const isJSFile = isSourceFileJS(declSourceFile);
             const call = tryCast(parent.parent, isCallExpression);
-            return { kind: InfoKind.ClassOrInterface, token, call, modifierFlags, parentDeclaration: classOrInterface, declSourceFile, isJSFile };
+            return { kind: InfoKind.TypeLikeDeclaration, token, call, modifierFlags, parentDeclaration: declaration, declSourceFile, isJSFile };
         }
 
         const enumDeclaration = find(symbol.declarations, isEnumDeclaration);
         if (enumDeclaration && !isPrivateIdentifier(token) && !isSourceFileFromLibrary(program, enumDeclaration.getSourceFile())) {
             return { kind: InfoKind.Enum, token, parentDeclaration: enumDeclaration };
         }
+
         return undefined;
     }
 
@@ -241,13 +245,13 @@ namespace ts.codefix {
         return program.isSourceFileFromExternalLibrary(node) || program.isSourceFileDefaultLibrary(node);
     }
 
-    function getActionsForMissingMemberDeclaration(context: CodeFixContext, info: ClassOrInterfaceInfo): CodeFixAction[] | undefined {
+    function getActionsForMissingMemberDeclaration(context: CodeFixContext, info: TypeLikeDeclarationInfo): CodeFixAction[] | undefined {
         return info.isJSFile ? singleElementArray(createActionForAddMissingMemberInJavascriptFile(context, info)) :
             createActionsForAddMissingMemberInTypeScriptFile(context, info);
     }
 
-    function createActionForAddMissingMemberInJavascriptFile(context: CodeFixContext, { parentDeclaration, declSourceFile, modifierFlags, token }: ClassOrInterfaceInfo): CodeFixAction | undefined {
-        if (isInterfaceDeclaration(parentDeclaration)) {
+    function createActionForAddMissingMemberInJavascriptFile(context: CodeFixContext, { parentDeclaration, declSourceFile, modifierFlags, token }: TypeLikeDeclarationInfo): CodeFixAction | undefined {
+        if (isInterfaceDeclaration(parentDeclaration) || isTypeLiteralNode(parentDeclaration)) {
             return undefined;
         }
 
@@ -262,7 +266,7 @@ namespace ts.codefix {
         return createCodeFixAction(fixMissingMember, changes, [diagnostic, token.text], fixMissingMember, Diagnostics.Add_all_missing_members);
     }
 
-    function addMissingMemberInJs(changeTracker: textChanges.ChangeTracker, declSourceFile: SourceFile, classDeclaration: ClassLikeDeclaration, token: Identifier | PrivateIdentifier, makeStatic: boolean): void {
+    function addMissingMemberInJs(changeTracker: textChanges.ChangeTracker, sourceFile: SourceFile, classDeclaration: ClassLikeDeclaration, token: Identifier | PrivateIdentifier, makeStatic: boolean): void {
         const tokenName = token.text;
         if (makeStatic) {
             if (classDeclaration.kind === SyntaxKind.ClassExpression) {
@@ -270,11 +274,10 @@ namespace ts.codefix {
             }
             const className = classDeclaration.name!.getText();
             const staticInitialization = initializePropertyToUndefined(factory.createIdentifier(className), tokenName);
-            changeTracker.insertNodeAfter(declSourceFile, classDeclaration, staticInitialization);
+            changeTracker.insertNodeAfter(sourceFile, classDeclaration, staticInitialization);
         }
         else if (isPrivateIdentifier(token)) {
             const property = factory.createPropertyDeclaration(
-                /*decorators*/ undefined,
                 /*modifiers*/ undefined,
                 tokenName,
                 /*questionToken*/ undefined,
@@ -283,10 +286,10 @@ namespace ts.codefix {
 
             const lastProp = getNodeToInsertPropertyAfter(classDeclaration);
             if (lastProp) {
-                changeTracker.insertNodeAfter(declSourceFile, lastProp, property);
+                changeTracker.insertNodeAfter(sourceFile, lastProp, property);
             }
             else {
-                changeTracker.insertNodeAtClassStart(declSourceFile, classDeclaration, property);
+                changeTracker.insertMemberAtStart(sourceFile, classDeclaration, property);
             }
         }
         else {
@@ -295,7 +298,7 @@ namespace ts.codefix {
                 return;
             }
             const propertyInitialization = initializePropertyToUndefined(factory.createThis(), tokenName);
-            changeTracker.insertNodeAtConstructorEnd(declSourceFile, classConstructor, propertyInitialization);
+            changeTracker.insertNodeAtConstructorEnd(sourceFile, classConstructor, propertyInitialization);
         }
     }
 
@@ -303,7 +306,7 @@ namespace ts.codefix {
         return factory.createExpressionStatement(factory.createAssignment(factory.createPropertyAccessExpression(obj, propertyName), createUndefined()));
     }
 
-    function createActionsForAddMissingMemberInTypeScriptFile(context: CodeFixContext, { parentDeclaration, declSourceFile, modifierFlags, token }: ClassOrInterfaceInfo): CodeFixAction[] | undefined {
+    function createActionsForAddMissingMemberInTypeScriptFile(context: CodeFixContext, { parentDeclaration, declSourceFile, modifierFlags, token }: TypeLikeDeclarationInfo): CodeFixAction[] | undefined {
         const memberName = token.text;
         const isStatic = modifierFlags & ModifierFlags.Static;
         const typeNode = getTypeNode(context.program.getTypeChecker(), parentDeclaration, token);
@@ -322,13 +325,13 @@ namespace ts.codefix {
         return actions;
     }
 
-    function getTypeNode(checker: TypeChecker, classDeclaration: ClassOrInterface, token: Node) {
+    function getTypeNode(checker: TypeChecker, node: ClassLikeDeclaration | InterfaceDeclaration | TypeLiteralNode, token: Node) {
         let typeNode: TypeNode | undefined;
         if (token.parent.parent.kind === SyntaxKind.BinaryExpression) {
             const binaryExpression = token.parent.parent as BinaryExpression;
             const otherExpression = token.parent === binaryExpression.left ? binaryExpression.right : binaryExpression.left;
             const widenedType = checker.getWidenedType(checker.getBaseTypeOfLiteralType(checker.getTypeAtLocation(otherExpression)));
-            typeNode = checker.typeToTypeNode(widenedType, classDeclaration, NodeBuilderFlags.NoTruncation);
+            typeNode = checker.typeToTypeNode(widenedType, node, NodeBuilderFlags.NoTruncation);
         }
         else {
             const contextualType = checker.getContextualType(token.parent as Expression);
@@ -337,39 +340,36 @@ namespace ts.codefix {
         return typeNode || factory.createKeywordTypeNode(SyntaxKind.AnyKeyword);
     }
 
-    function addPropertyDeclaration(changeTracker: textChanges.ChangeTracker, declSourceFile: SourceFile, classDeclaration: ClassOrInterface, tokenName: string, typeNode: TypeNode, modifierFlags: ModifierFlags): void {
-        const property = factory.createPropertyDeclaration(
-            /*decorators*/ undefined,
-            /*modifiers*/ modifierFlags ? factory.createNodeArray(factory.createModifiersFromModifierFlags(modifierFlags)) : undefined,
-            tokenName,
-            /*questionToken*/ undefined,
-            typeNode,
-            /*initializer*/ undefined);
+    function addPropertyDeclaration(changeTracker: textChanges.ChangeTracker, sourceFile: SourceFile, node: ClassLikeDeclaration | InterfaceDeclaration | TypeLiteralNode, tokenName: string, typeNode: TypeNode, modifierFlags: ModifierFlags): void {
+        const modifiers = modifierFlags ? factory.createNodeArray(factory.createModifiersFromModifierFlags(modifierFlags)) : undefined;
 
-        const lastProp = getNodeToInsertPropertyAfter(classDeclaration);
+        const property = isClassLike(node)
+            ? factory.createPropertyDeclaration(modifiers, tokenName, /*questionToken*/ undefined, typeNode, /*initializer*/ undefined)
+            : factory.createPropertySignature(/*modifiers*/ undefined, tokenName, /*questionToken*/ undefined, typeNode);
+
+        const lastProp = getNodeToInsertPropertyAfter(node);
         if (lastProp) {
-            changeTracker.insertNodeAfter(declSourceFile, lastProp, property);
+            changeTracker.insertNodeAfter(sourceFile, lastProp, property);
         }
         else {
-            changeTracker.insertNodeAtClassStart(declSourceFile, classDeclaration, property);
+            changeTracker.insertMemberAtStart(sourceFile, node, property);
         }
     }
 
     // Gets the last of the first run of PropertyDeclarations, or undefined if the class does not start with a PropertyDeclaration.
-    function getNodeToInsertPropertyAfter(cls: ClassOrInterface): PropertyDeclaration | undefined {
+    function getNodeToInsertPropertyAfter(node: ClassLikeDeclaration | InterfaceDeclaration | TypeLiteralNode): PropertyDeclaration | undefined {
         let res: PropertyDeclaration | undefined;
-        for (const member of cls.members) {
+        for (const member of node.members) {
             if (!isPropertyDeclaration(member)) break;
             res = member;
         }
         return res;
     }
 
-    function createAddIndexSignatureAction(context: CodeFixContext, declSourceFile: SourceFile, classDeclaration: ClassOrInterface, tokenName: string, typeNode: TypeNode): CodeFixAction {
+    function createAddIndexSignatureAction(context: CodeFixContext, sourceFile: SourceFile, node: ClassLikeDeclaration | InterfaceDeclaration | TypeLiteralNode, tokenName: string, typeNode: TypeNode): CodeFixAction {
         // Index signatures cannot have the static modifier.
         const stringTypeNode = factory.createKeywordTypeNode(SyntaxKind.StringKeyword);
         const indexingParameter = factory.createParameterDeclaration(
-            /*decorators*/ undefined,
             /*modifiers*/ undefined,
             /*dotDotDotToken*/ undefined,
             "x",
@@ -377,17 +377,16 @@ namespace ts.codefix {
             stringTypeNode,
             /*initializer*/ undefined);
         const indexSignature = factory.createIndexSignature(
-            /*decorators*/ undefined,
             /*modifiers*/ undefined,
             [indexingParameter],
             typeNode);
 
-        const changes = textChanges.ChangeTracker.with(context, t => t.insertNodeAtClassStart(declSourceFile, classDeclaration, indexSignature));
+        const changes = textChanges.ChangeTracker.with(context, t => t.insertMemberAtStart(sourceFile, node, indexSignature));
         // No fixId here because code-fix-all currently only works on adding individual named properties.
         return createCodeFixActionWithoutFixAll(fixMissingMember, changes, [Diagnostics.Add_index_signature_for_property_0, tokenName]);
     }
 
-    function getActionsForMissingMethodDeclaration(context: CodeFixContext, info: ClassOrInterfaceInfo): CodeFixAction[] | undefined {
+    function getActionsForMissingMethodDeclaration(context: CodeFixContext, info: TypeLikeDeclarationInfo): CodeFixAction[] | undefined {
         const { parentDeclaration, declSourceFile, modifierFlags, token, call } = info;
         if (call === undefined) {
             return undefined;
@@ -413,17 +412,18 @@ namespace ts.codefix {
         callExpression: CallExpression,
         name: Identifier,
         modifierFlags: ModifierFlags,
-        parentDeclaration: ClassOrInterface,
+        parentDeclaration: ClassLikeDeclaration | InterfaceDeclaration | TypeLiteralNode,
         sourceFile: SourceFile,
     ): void {
         const importAdder = createImportAdder(sourceFile, context.program, context.preferences, context.host);
-        const methodDeclaration = createSignatureDeclarationFromCallExpression(SyntaxKind.MethodDeclaration, context, importAdder, callExpression, name, modifierFlags, parentDeclaration) as MethodDeclaration;
-        const containingMethodDeclaration = findAncestor(callExpression, n => isMethodDeclaration(n) || isConstructorDeclaration(n));
-        if (containingMethodDeclaration && containingMethodDeclaration.parent === parentDeclaration) {
-            changes.insertNodeAfter(sourceFile, containingMethodDeclaration, methodDeclaration);
+        const kind = isClassLike(parentDeclaration) ? SyntaxKind.MethodDeclaration : SyntaxKind.MethodSignature;
+        const signatureDeclaration = createSignatureDeclarationFromCallExpression(kind, context, importAdder, callExpression, name, modifierFlags, parentDeclaration) as MethodDeclaration;
+        const containingMethodDeclaration = tryGetContainingMethodDeclaration(parentDeclaration, callExpression);
+        if (containingMethodDeclaration) {
+            changes.insertNodeAfter(sourceFile, containingMethodDeclaration, signatureDeclaration);
         }
         else {
-            changes.insertNodeAtClassStart(sourceFile, parentDeclaration, methodDeclaration);
+            changes.insertMemberAtStart(sourceFile, parentDeclaration, signatureDeclaration);
         }
         importAdder.writeFixes(changes);
     }
@@ -442,7 +442,6 @@ namespace ts.codefix {
         const enumMember = factory.createEnumMember(token, hasStringInitializer ? factory.createStringLiteral(token.text) : undefined);
         changes.replaceNode(parentDeclaration.getSourceFile(), parentDeclaration, factory.updateEnumDeclaration(
             parentDeclaration,
-            parentDeclaration.decorators,
             parentDeclaration.modifiers,
             parentDeclaration.name,
             concatenate(parentDeclaration.members, singleElementArray(enumMember))
@@ -456,6 +455,7 @@ namespace ts.codefix {
         const importAdder = createImportAdder(context.sourceFile, context.program, context.preferences, context.host);
         const functionDeclaration = createSignatureDeclarationFromCallExpression(SyntaxKind.FunctionDeclaration, context, importAdder, info.call, idText(info.token), info.modifierFlags, info.parentDeclaration) as FunctionDeclaration;
         changes.insertNodeAtEndOfScope(info.sourceFile, info.parentDeclaration, functionDeclaration);
+        importAdder.writeFixes(changes);
     }
 
     function addJsxAttributes(changes: textChanges.ChangeTracker, context: CodeFixContextBase, info: JsxAttributesInfo) {
@@ -465,21 +465,27 @@ namespace ts.codefix {
         const jsxAttributesNode = info.parentDeclaration.attributes;
         const hasSpreadAttribute = some(jsxAttributesNode.properties, isJsxSpreadAttribute);
         const attrs = map(info.attributes, attr => {
-            const value = attr.valueDeclaration ? tryGetValueFromType(context, checker, importAdder, quotePreference, checker.getTypeAtLocation(attr.valueDeclaration)) : createUndefined();
-            return factory.createJsxAttribute(factory.createIdentifier(attr.name), factory.createJsxExpression(/*dotDotDotToken*/ undefined, value));
+            const value = tryGetValueFromType(context, checker, importAdder, quotePreference, checker.getTypeOfSymbol(attr), info.parentDeclaration);
+            const name = factory.createIdentifier(attr.name);
+            const jsxAttribute = factory.createJsxAttribute(name, factory.createJsxExpression(/*dotDotDotToken*/ undefined, value));
+            // formattingScanner requires the Identifier to have a context for scanning attributes with "-" (data-foo).
+            setParent(name, jsxAttribute);
+            return jsxAttribute;
         });
         const jsxAttributes = factory.createJsxAttributes(hasSpreadAttribute ? [...attrs, ...jsxAttributesNode.properties] : [...jsxAttributesNode.properties, ...attrs]);
         const options = { prefix: jsxAttributesNode.pos === jsxAttributesNode.end ? " " : undefined };
         changes.replaceNode(context.sourceFile, jsxAttributesNode, jsxAttributes, options);
+        importAdder.writeFixes(changes);
     }
 
     function addObjectLiteralProperties(changes: textChanges.ChangeTracker, context: CodeFixContextBase, info: ObjectLiteralInfo) {
         const importAdder = createImportAdder(context.sourceFile, context.program, context.preferences, context.host);
         const quotePreference = getQuotePreference(context.sourceFile, context.preferences);
+        const target = getEmitScriptTarget(context.program.getCompilerOptions());
         const checker = context.program.getTypeChecker();
         const props = map(info.properties, prop => {
-            const initializer = prop.valueDeclaration ? tryGetValueFromType(context, checker, importAdder, quotePreference, checker.getTypeAtLocation(prop.valueDeclaration)) : createUndefined();
-            return factory.createPropertyAssignment(prop.name, initializer);
+            const initializer = tryGetValueFromType(context, checker, importAdder, quotePreference, checker.getTypeOfSymbol(prop), info.parentDeclaration);
+            return factory.createPropertyAssignment(createPropertyNameFromSymbol(prop, target, quotePreference, checker), initializer);
         });
         const options = {
             leadingTriviaOption: textChanges.LeadingTriviaOption.Exclude,
@@ -487,9 +493,10 @@ namespace ts.codefix {
             indentation: info.indentation
         };
         changes.replaceNode(context.sourceFile, info.parentDeclaration, factory.createObjectLiteralExpression([...info.parentDeclaration.properties, ...props], /*multiLine*/ true), options);
+        importAdder.writeFixes(changes);
     }
 
-    function tryGetValueFromType(context: CodeFixContextBase, checker: TypeChecker, importAdder: ImportAdder, quotePreference: QuotePreference, type: Type): Expression {
+    function tryGetValueFromType(context: CodeFixContextBase, checker: TypeChecker, importAdder: ImportAdder, quotePreference: QuotePreference, type: Type, enclosingDeclaration: Node | undefined): Expression {
         if (type.flags & TypeFlags.AnyOrUnknown) {
             return createUndefined();
         }
@@ -526,7 +533,7 @@ namespace ts.codefix {
             return factory.createNull();
         }
         if (type.flags & TypeFlags.Union) {
-            const expression = firstDefined((type as UnionType).types, t => tryGetValueFromType(context, checker, importAdder, quotePreference, t));
+            const expression = firstDefined((type as UnionType).types, t => tryGetValueFromType(context, checker, importAdder, quotePreference, t, enclosingDeclaration));
             return expression ?? createUndefined();
         }
         if (checker.isArrayLikeType(type)) {
@@ -534,7 +541,7 @@ namespace ts.codefix {
         }
         if (isObjectLiteralType(type)) {
             const props = map(checker.getPropertiesOfType(type), prop => {
-                const initializer = prop.valueDeclaration ? tryGetValueFromType(context, checker, importAdder, quotePreference, checker.getTypeAtLocation(prop.valueDeclaration)) : createUndefined();
+                const initializer = prop.valueDeclaration ? tryGetValueFromType(context, checker, importAdder, quotePreference, checker.getTypeAtLocation(prop.valueDeclaration), enclosingDeclaration) : createUndefined();
                 return factory.createPropertyAssignment(prop.name, initializer);
             });
             return factory.createObjectLiteralExpression(props, /*multiLine*/ true);
@@ -547,7 +554,7 @@ namespace ts.codefix {
             if (signature === undefined) return createUndefined();
 
             const func = createSignatureDeclarationFromSignature(SyntaxKind.FunctionExpression, context, quotePreference, signature[0],
-                createStubbedBody(Diagnostics.Function_not_implemented.message, quotePreference), /*name*/ undefined, /*modifiers*/ undefined, /*optional*/ undefined, /*enclosingDeclaration*/ undefined, importAdder) as FunctionExpression | undefined;
+                createStubbedBody(Diagnostics.Function_not_implemented.message, quotePreference), /*name*/ undefined, /*modifiers*/ undefined, /*optional*/ undefined, /*enclosingDeclaration*/ enclosingDeclaration, importAdder) as FunctionExpression | undefined;
             return func ?? createUndefined();
         }
         if (getObjectFlags(type) & ObjectFlags.Class) {
@@ -571,7 +578,7 @@ namespace ts.codefix {
             ((getObjectFlags(type) & ObjectFlags.ObjectLiteral) || (type.symbol && tryCast(singleOrUndefined(type.symbol.declarations), isTypeLiteralNode)));
     }
 
-    function getUnmatchedAttributes(checker: TypeChecker, source: JsxOpeningLikeElement) {
+    function getUnmatchedAttributes(checker: TypeChecker, target: ScriptTarget, source: JsxOpeningLikeElement) {
         const attrsType = checker.getContextualType(source.attributes);
         if (attrsType === undefined) return emptyArray;
 
@@ -591,6 +598,24 @@ namespace ts.codefix {
             }
         }
         return filter(targetProps, targetProp =>
-            !((targetProp.flags & SymbolFlags.Optional || getCheckFlags(targetProp) & CheckFlags.Partial) || seenNames.has(targetProp.escapedName)));
+            isIdentifierText(targetProp.name, target, LanguageVariant.JSX) && !((targetProp.flags & SymbolFlags.Optional || getCheckFlags(targetProp) & CheckFlags.Partial) || seenNames.has(targetProp.escapedName)));
+    }
+
+    function tryGetContainingMethodDeclaration(node: ClassLikeDeclaration | InterfaceDeclaration | TypeLiteralNode, callExpression: CallExpression) {
+        if (isTypeLiteralNode(node)) {
+            return undefined;
+        }
+        const declaration = findAncestor(callExpression, n => isMethodDeclaration(n) || isConstructorDeclaration(n));
+        return declaration && declaration.parent === node ? declaration : undefined;
+    }
+
+    function createPropertyNameFromSymbol(symbol: Symbol, target: ScriptTarget, quotePreference: QuotePreference, checker: TypeChecker) {
+        if (isTransientSymbol(symbol) && symbol.nameType && symbol.nameType.flags & TypeFlags.UniqueESSymbol) {
+            const expression = checker.symbolToExpression((symbol.nameType as UniqueESSymbolType).symbol, SymbolFlags.Value, symbol.valueDeclaration, NodeBuilderFlags.AllowUniqueESSymbolType);
+            if (expression) {
+                return factory.createComputedPropertyName(expression);
+            }
+        }
+        return createPropertyNameNodeForIdentifierOrLiteral(symbol.name, target, quotePreference === QuotePreference.Single);
     }
 }
