@@ -87,7 +87,6 @@ namespace ts.codefix {
                     }
                 }
                 addClassElement(factory.createPropertyDeclaration(
-                    /*decorators*/ undefined,
                     modifiers,
                     name,
                     optional && (preserveOptional & PreserveOptionalFlags.Property) ? factory.createToken(SyntaxKind.QuestionToken) : undefined,
@@ -111,7 +110,6 @@ namespace ts.codefix {
                 for (const accessor of orderedAccessors) {
                     if (isGetAccessorDeclaration(accessor)) {
                         addClassElement(factory.createGetAccessorDeclaration(
-                            /*decorators*/ undefined,
                             modifiers,
                             name,
                             emptyArray,
@@ -123,7 +121,6 @@ namespace ts.codefix {
                         const parameter = getSetAccessorValueParameter(accessor);
                         const parameterName = parameter && isIdentifier(parameter.name) ? idText(parameter.name) : undefined;
                         addClassElement(factory.createSetAccessorDeclaration(
-                            /*decorators*/ undefined,
                             modifiers,
                             name,
                             createDummyParameters(1, [parameterName], [typeNode], 1, /*inJs*/ false),
@@ -141,7 +138,7 @@ namespace ts.codefix {
                 // If there is more than one overload but no implementation signature
                 // (eg: an abstract method or interface declaration), there is a 1-1
                 // correspondence of declarations and signatures.
-                const signatures = checker.getSignaturesOfType(type, SignatureKind.Call);
+                const signatures = type.isUnion() ? flatMap(type.types, t => t.getCallSignatures()) : type.getCallSignatures();
                 if (!some(signatures)) {
                     break;
                 }
@@ -192,7 +189,11 @@ namespace ts.codefix {
         const program = context.program;
         const checker = program.getTypeChecker();
         const scriptTarget = getEmitScriptTarget(program.getCompilerOptions());
-        const flags = NodeBuilderFlags.NoTruncation | NodeBuilderFlags.NoUndefinedOptionalParameterType | NodeBuilderFlags.SuppressAnyReturnType | (quotePreference === QuotePreference.Single ? NodeBuilderFlags.UseSingleQuotesForStringLiteralType : 0);
+        const flags =
+            NodeBuilderFlags.NoTruncation
+            | NodeBuilderFlags.SuppressAnyReturnType
+            | NodeBuilderFlags.AllowEmptyTuple
+            | (quotePreference === QuotePreference.Single ? NodeBuilderFlags.UseSingleQuotesForStringLiteralType : NodeBuilderFlags.None);
         const signatureDeclaration = checker.signatureToSignatureDeclaration(signature, kind, enclosingDeclaration, flags, getNoopSymbolTrackerWithResolver(context)) as ArrowFunction | FunctionExpression | MethodDeclaration;
         if (!signatureDeclaration) {
             return undefined;
@@ -222,6 +223,7 @@ namespace ts.codefix {
                     }
                     return factory.updateTypeParameterDeclaration(
                         typeParameterDecl,
+                        typeParameterDecl.modifiers,
                         typeParameterDecl.name,
                         constraint,
                         defaultType
@@ -240,7 +242,6 @@ namespace ts.codefix {
                 }
                 return factory.updateParameterDeclaration(
                     parameterDecl,
-                    parameterDecl.decorators,
                     parameterDecl.modifiers,
                     parameterDecl.dotDotDotToken,
                     parameterDecl.name,
@@ -270,13 +271,13 @@ namespace ts.codefix {
             return factory.updateArrowFunction(signatureDeclaration, modifiers, typeParameters, parameters, type, signatureDeclaration.equalsGreaterThanToken, body ?? signatureDeclaration.body);
         }
         if (isMethodDeclaration(signatureDeclaration)) {
-            return factory.updateMethodDeclaration(signatureDeclaration, /* decorators */ undefined, modifiers, asteriskToken, name ?? factory.createIdentifier(""), questionToken, typeParameters, parameters, type, body);
+            return factory.updateMethodDeclaration(signatureDeclaration, modifiers, asteriskToken, name ?? factory.createIdentifier(""), questionToken, typeParameters, parameters, type, body);
         }
         return undefined;
     }
 
     export function createSignatureDeclarationFromCallExpression(
-        kind: SyntaxKind.MethodDeclaration | SyntaxKind.FunctionDeclaration,
+        kind: SyntaxKind.MethodDeclaration | SyntaxKind.FunctionDeclaration | SyntaxKind.MethodSignature,
         context: CodeFixContextBase,
         importAdder: ImportAdder,
         call: CallExpression,
@@ -294,8 +295,10 @@ namespace ts.codefix {
         const contextualType = isJs ? undefined : checker.getContextualType(call);
         const names = map(args, arg =>
             isIdentifier(arg) ? arg.text : isPropertyAccessExpression(arg) && isIdentifier(arg.name) ? arg.name.text : undefined);
-        const types = isJs ? [] : map(args, arg =>
-            typeToAutoImportableTypeNode(checker, importAdder, checker.getBaseTypeOfLiteralType(checker.getTypeAtLocation(arg)), contextNode, scriptTarget, /*flags*/ undefined, tracker));
+        const instanceTypes = isJs ? [] : map(args, arg => checker.getTypeAtLocation(arg));
+        const { argumentTypeNodes, argumentTypeParameters } = getArgumentTypesAndTypeParameters(
+            checker, importAdder, instanceTypes, contextNode, scriptTarget, /*flags*/ undefined, tracker
+        );
 
         const modifiers = modifierFlags
             ? factory.createNodeArray(factory.createModifiersFromModifierFlags(modifierFlags))
@@ -303,40 +306,77 @@ namespace ts.codefix {
         const asteriskToken = isYieldExpression(parent)
             ? factory.createToken(SyntaxKind.AsteriskToken)
             : undefined;
-        const typeParameters = isJs || typeArguments === undefined
-            ? undefined
-            : map(typeArguments, (_, i) =>
-                factory.createTypeParameterDeclaration(CharacterCodes.T + typeArguments.length - 1 <= CharacterCodes.Z ? String.fromCharCode(CharacterCodes.T + i) : `T${i}`));
-        const parameters = createDummyParameters(args.length, names, types, /*minArgumentCount*/ undefined, isJs);
+        const typeParameters = isJs ? undefined : createTypeParametersForArguments(checker, argumentTypeParameters, typeArguments);
+        const parameters = createDummyParameters(args.length, names, argumentTypeNodes, /*minArgumentCount*/ undefined, isJs);
         const type = isJs || contextualType === undefined
             ? undefined
             : checker.typeToTypeNode(contextualType, contextNode, /*flags*/ undefined, tracker);
 
-        if (kind === SyntaxKind.MethodDeclaration) {
-            return factory.createMethodDeclaration(
-                /*decorators*/ undefined,
-                modifiers,
-                asteriskToken,
-                name,
-                /*questionToken*/ undefined,
-                typeParameters,
-                parameters,
-                type,
-                isInterfaceDeclaration(contextNode) ? undefined : createStubbedMethodBody(quotePreference)
-            );
+        switch (kind) {
+            case SyntaxKind.MethodDeclaration:
+                return factory.createMethodDeclaration(
+                    modifiers,
+                    asteriskToken,
+                    name,
+                    /*questionToken*/ undefined,
+                    typeParameters,
+                    parameters,
+                    type,
+                    createStubbedMethodBody(quotePreference)
+                );
+            case SyntaxKind.MethodSignature:
+                return factory.createMethodSignature(
+                    modifiers,
+                    name,
+                    /*questionToken*/ undefined,
+                    typeParameters,
+                    parameters,
+                    type === undefined ? factory.createKeywordTypeNode(SyntaxKind.UnknownKeyword) : type
+                );
+            case SyntaxKind.FunctionDeclaration:
+                return factory.createFunctionDeclaration(
+                    modifiers,
+                    asteriskToken,
+                    name,
+                    typeParameters,
+                    parameters,
+                    type,
+                    createStubbedBody(Diagnostics.Function_not_implemented.message, quotePreference)
+                );
+            default:
+                Debug.fail("Unexpected kind");
         }
-        return factory.createFunctionDeclaration(
-            /*decorators*/ undefined,
-            modifiers,
-            asteriskToken,
-            name,
-            typeParameters,
-            parameters,
-            type,
-            createStubbedBody(Diagnostics.Function_not_implemented.message, quotePreference)
+    }
+
+    interface ArgumentTypeParameterAndConstraint {
+        argumentType: Type;
+        constraint?: TypeNode;
+    }
+
+    function createTypeParametersForArguments(checker: TypeChecker, argumentTypeParameters: [string, ArgumentTypeParameterAndConstraint | undefined][], typeArguments: NodeArray<TypeNode> | undefined) {
+        const usedNames = new Set(argumentTypeParameters.map(pair => pair[0]));
+        const constraintsByName = new Map(argumentTypeParameters);
+
+        if (typeArguments) {
+            const typeArgumentsWithNewTypes = typeArguments.filter(typeArgument => !argumentTypeParameters.some(pair => checker.getTypeAtLocation(typeArgument) === pair[1]?.argumentType));
+            const targetSize = usedNames.size + typeArgumentsWithNewTypes.length;
+            for (let i = 0; usedNames.size < targetSize; i += 1) {
+                usedNames.add(createTypeParameterName(i));
+            }
+        }
+
+        return map(
+            arrayFrom(usedNames.values()),
+            usedName => factory.createTypeParameterDeclaration(/*modifiers*/ undefined, usedName, constraintsByName.get(usedName)?.constraint),
         );
     }
 
+    function createTypeParameterName(index: number) {
+        return CharacterCodes.T + index <= CharacterCodes.Z
+            ? String.fromCharCode(CharacterCodes.T + index)
+            : `T${index}`;
+    }
+
     export function typeToAutoImportableTypeNode(checker: TypeChecker, importAdder: ImportAdder, type: Type, contextNode: Node | undefined, scriptTarget: ScriptTarget, flags?: NodeBuilderFlags, tracker?: SymbolTracker): TypeNode | undefined {
         let typeNode = checker.typeToTypeNode(type, contextNode, flags, tracker);
         if (typeNode && isImportTypeNode(typeNode)) {
@@ -346,20 +386,124 @@ namespace ts.codefix {
                 typeNode = importableReference.typeNode;
             }
         }
+
         // Ensure nodes are fresh so they can have different positions when going through formatting.
         return getSynthesizedDeepClone(typeNode);
     }
 
+    function typeContainsTypeParameter(type: Type) {
+        if (type.isUnionOrIntersection()) {
+            return type.types.some(typeContainsTypeParameter);
+        }
+
+        return type.flags & TypeFlags.TypeParameter;
+    }
+
+    export function getArgumentTypesAndTypeParameters(checker: TypeChecker, importAdder: ImportAdder, instanceTypes: Type[], contextNode: Node | undefined, scriptTarget: ScriptTarget, flags?: NodeBuilderFlags, tracker?: SymbolTracker) {
+        // Types to be used as the types of the parameters in the new function
+        // E.g. from this source:
+        //   added("", 0)
+        // The value will look like:
+        //   [{ typeName: { text: "string" } }, { typeName: { text: "number" }]
+        // And in the output function will generate:
+        //   function added(a: string, b: number) { ... }
+        const argumentTypeNodes: TypeNode[] = [];
+
+        // Names of type parameters provided as arguments to the call
+        // E.g. from this source:
+        //   added<T, U>(value);
+        // The value will look like:
+        //   [
+        //     ["T", { argumentType: { typeName: { text: "T" } } } ],
+        //     ["U", { argumentType: { typeName: { text: "U" } } } ],
+        //   ]
+        // And in the output function will generate:
+        //   function added<T, U>() { ... }
+        const argumentTypeParameters = new Map<string, ArgumentTypeParameterAndConstraint | undefined>();
+
+        for (let i = 0; i < instanceTypes.length; i += 1) {
+            const instanceType = instanceTypes[i];
+
+            // If the instance type contains a deep reference to an existing type parameter,
+            // instead of copying the full union or intersection, create a new type parameter
+            // E.g. from this source:
+            //   function existing<T, U>(value: T | U & string) {
+            //     added/*1*/(value);
+            // We don't want to output this:
+            //    function added<T>(value: T | U & string) { ... }
+            // We instead want to output:
+            //    function added<T>(value: T) { ... }
+            if (instanceType.isUnionOrIntersection() && instanceType.types.some(typeContainsTypeParameter)) {
+                const synthesizedTypeParameterName = createTypeParameterName(i);
+                argumentTypeNodes.push(factory.createTypeReferenceNode(synthesizedTypeParameterName));
+                argumentTypeParameters.set(synthesizedTypeParameterName, undefined);
+                continue;
+            }
+
+            // Widen the type so we don't emit nonsense annotations like "function fn(x: 3) {"
+            const widenedInstanceType = checker.getBaseTypeOfLiteralType(instanceType);
+            const argumentTypeNode = typeToAutoImportableTypeNode(checker, importAdder, widenedInstanceType, contextNode, scriptTarget, flags, tracker);
+            if (!argumentTypeNode) {
+                continue;
+            }
+
+            argumentTypeNodes.push(argumentTypeNode);
+            const argumentTypeParameter = getFirstTypeParameterName(instanceType);
+
+            // If the instance type is a type parameter with a constraint (other than an anonymous object),
+            // remember that constraint for when we create the new type parameter
+            // E.g. from this source:
+            //   function existing<T extends string>(value: T) {
+            //     added/*1*/(value);
+            // We don't want to output this:
+            //    function added<T>(value: T) { ... }
+            // We instead want to output:
+            //    function added<T extends string>(value: T) { ... }
+            const instanceTypeConstraint = instanceType.isTypeParameter() && instanceType.constraint && !isAnonymousObjectConstraintType(instanceType.constraint)
+                ? typeToAutoImportableTypeNode(checker, importAdder, instanceType.constraint, contextNode, scriptTarget, flags, tracker)
+                : undefined;
+
+            if (argumentTypeParameter) {
+                argumentTypeParameters.set(argumentTypeParameter, { argumentType: instanceType, constraint: instanceTypeConstraint });
+            }
+        }
+
+        return { argumentTypeNodes, argumentTypeParameters: arrayFrom(argumentTypeParameters.entries()) };
+    }
+
+    function isAnonymousObjectConstraintType(type: Type) {
+        return (type.flags & TypeFlags.Object) && (type as ObjectType).objectFlags === ObjectFlags.Anonymous;
+    }
+
+    function getFirstTypeParameterName(type: Type): string | undefined {
+        if (type.flags & (TypeFlags.Union | TypeFlags.Intersection)) {
+            for (const subType of (type as UnionType | IntersectionType).types) {
+                const subTypeName = getFirstTypeParameterName(subType);
+                if (subTypeName) {
+                    return subTypeName;
+                }
+            }
+        }
+
+        return type.flags & TypeFlags.TypeParameter
+            ? type.getSymbol()?.getName()
+            : undefined;
+    }
+
     function createDummyParameters(argCount: number, names: (string | undefined)[] | undefined, types: (TypeNode | undefined)[] | undefined, minArgumentCount: number | undefined, inJs: boolean): ParameterDeclaration[] {
         const parameters: ParameterDeclaration[] = [];
+        const parameterNameCounts = new Map<string, number>();
         for (let i = 0; i < argCount; i++) {
+            const parameterName = names?.[i] || `arg${i}`;
+            const parameterNameCount = parameterNameCounts.get(parameterName);
+            parameterNameCounts.set(parameterName, (parameterNameCount || 0) + 1);
+
             const newParameter = factory.createParameterDeclaration(
-                /*decorators*/ undefined,
                 /*modifiers*/ undefined,
                 /*dotDotDotToken*/ undefined,
-                /*name*/ names && names[i] || `arg${i}`,
+                /*name*/ parameterName + (parameterNameCount || ""),
                 /*questionToken*/ minArgumentCount !== undefined && i >= minArgumentCount ? factory.createToken(SyntaxKind.QuestionToken) : undefined,
-                /*type*/ inJs ? undefined : types && types[i] || factory.createKeywordTypeNode(SyntaxKind.AnyKeyword),
+                /*type*/ inJs ? undefined : types?.[i] || factory.createKeywordTypeNode(SyntaxKind.UnknownKeyword),
                 /*initializer*/ undefined);
             parameters.push(newParameter);
         }
@@ -398,14 +542,12 @@ namespace ts.codefix {
         const parameters = createDummyParameters(maxNonRestArgs, maxArgsParameterSymbolNames, /* types */ undefined, minArgumentCount, /*inJs*/ false);
 
         if (someSigHasRestParameter) {
-            const anyArrayType = factory.createArrayTypeNode(factory.createKeywordTypeNode(SyntaxKind.AnyKeyword));
             const restParameter = factory.createParameterDeclaration(
-                /*decorators*/ undefined,
                 /*modifiers*/ undefined,
                 factory.createToken(SyntaxKind.DotDotDotToken),
                 maxArgsParameterSymbolNames[maxNonRestArgs] || "rest",
                 /*questionToken*/ maxNonRestArgs >= minArgumentCount ? factory.createToken(SyntaxKind.QuestionToken) : undefined,
-                anyArrayType,
+                factory.createArrayTypeNode(factory.createKeywordTypeNode(SyntaxKind.UnknownKeyword)),
                 /*initializer*/ undefined);
             parameters.push(restParameter);
         }
@@ -439,7 +581,6 @@ namespace ts.codefix {
         body: Block | undefined
     ): MethodDeclaration {
         return factory.createMethodDeclaration(
-            /*decorators*/ undefined,
             modifiers,
             /*asteriskToken*/ undefined,
             name,
