@@ -15,6 +15,6 @@
 
 const exact = completion.globalTypesPlus(["I", "C"]);
 verify.completions(
-    { marker: ["1", "2"], exact: ["T", ...exact] },
-    { marker: ["3", "4", "5"], exact: completion.globalTypesPlus(["I", "C", "T"]) },
+    { marker: ["1", "2"], unsorted: ["T", ...exact] },
+    { marker: ["3", "4", "5"], exact: completion.globalTypesPlus(["C", "I", "T"]) },
 );
